"""
Helper functions for generating Concrete UUIDs
"""

# Force 'import uuid' to import the Python standard library module
# named "uuid", and not the "concrete.uuid" module
from __future__ import absolute_import

import uuid as python_uuid

import concrete

import random


def generate_UUID():
    """Helper function for generating a Concrete UUID object
    """
    c_uuid = concrete.UUID()
    c_uuid.uuidString = str(python_uuid.uuid4())
    return c_uuid


def hex_to_bin(h):
    return int(h, 16)

def bin_to_hex(b, n=None):
    h = hex(b)[2:]
    if n is None:
        n = len(h)
    elif len(h) > n:
        raise ValueError('hex string "%s" is longer than %d chars' % (h, n))
    return ('0' * (n - len(h))) + h

def split_uuid(u):
    p = u.split('-')

    valid_input = (
        len(p) == 5
        and len(p[0]) == 8
        and len(p[1]) == 4
        and len(p[2]) == 4
        and len(p[3]) == 4
        and len(p[4]) == 12
    )
    if not valid_input:
        raise ValueError(
            'uuid "%s" is not of the form xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
            % u
        )

    xs = p[0] + p[1]
    ys = p[2] + p[3]
    zs = p[4]

    return (xs, ys, zs)

def join_uuid(xs, ys, zs):
    valid_input = (
        len(xs) == 12
        and len(ys) == 8
        and len(zs) == 12
    )
    if not valid_input:
        raise ValueError('uuid pieces do not have lengths 12, 8, 12')

    u = '-'.join((xs[:8], xs[8:], ys[:4], ys[4:], zs))

    return u

def generate_hex_unif(n):
    return ''.join(random.choice('abcdef0123456789') for i in xrange(n))

def generate_uuid_unif():
    return join_uuid(generate_hex_unif(12), generate_hex_unif(8), generate_hex_unif(12))

class _AnalyticUUIDGenerator(object):
    '''
    UUID generator for a given analytic in a given communication.
    '''

    def __init__(self, u):
        (self._xs, ys, zs) = split_uuid(u)
        self._ys = generate_hex_unif(len(ys))
        self._z = hex_to_bin(generate_hex_unif(len(zs)))
        self._z_len = len(zs)
        self._z_bound = 2**(4*len(zs))
        self.n = 0

    def __iter__(self):
        return self

    def next(self):
        '''
        Generate and return a new concrete UUID.
        StopIteration will never be raised.
        '''
        self._z = (self._z + 1) % self._z_bound
        self.n += 1
        c_uuid = concrete.UUID()
        c_uuid.uuidString = join_uuid(self._xs, self._ys, bin_to_hex(self._z, self._z_len))
        return c_uuid

class AnalyticUUIDGeneratorFactory(object):
    '''
    Factory for a compressible UUID generator.

    One factory should be created per communication, and a new generator
    should be created from that factory for each analytic processing the
    communication.  Usually each program represents a single analytic,
    so common usage is

        augf = AnalyticUUIDGeneratorFactory(comm)
        aug = augf.create()
        for <each annotation object created by this analytic>:
            annotation. = aug.next()
            <add annotation to communication>

    or if you're creating a new communication

        augf = AnalyticUUIDGeneratorFactory()
        aug = augf.create()
        comm = <create communication>
        comm.uuid = aug.next()
        for <each annotation object created by this analytic>:
            annotation. = aug.next()
            <add annotation to communication>

    where the annotation objects might be objects of type
    Parse, DependencyParse, TokenTagging, CommunicationTagging, etc.
    '''

    def __init__(self, comm=None):
        if comm is None:
            self.comm_uuid = generate_uuid_unif()
        else:
            self.comm_uuid = comm.uuid.uuidString

    def create(self):
        '''
        Create and return a UUID generator for a new analytic.
        '''
        return _AnalyticUUIDGenerator(self.comm_uuid)



"""Create a simple (valid) Communication suitable for testing purposes
"""

import time
import tempfile
import os

from concrete import (
    AnnotationMetadata,
    Communication,
    Section,
    Sentence,
    TextSpan,
    Token,
    Tokenization,
    TokenizationKind,
    TokenList
)
from concrete.util.concrete_uuid import AnalyticUUIDGeneratorFactory
from concrete.util.file_io import CommunicationWriter


def create_simple_comm(comm_id, sentence_string="Super simple sentence ."):
    """Create a simple (valid) Communication suitable for testing purposes

    The Communication will have a single Section containing a single
    Sentence.

    Args:

    - `comm_id`: A string specifying a Communication ID
    - `sentence_string`: A string to be used for the sentence text.
       The string will be whitespace-tokenized.

    Returns:

    - A Concrete Communication object
    """
    toolname = "TEST"
    timestamp = int(time.time())

    augf = AnalyticUUIDGeneratorFactory()
    aug = augf.create()

    comm = Communication(
        id=comm_id,
        metadata=AnnotationMetadata(tool=toolname, timestamp=timestamp),
        type=toolname,
        uuid=aug.next()
    )

    tokenization = Tokenization(
        kind=TokenizationKind.TOKEN_LIST,
        metadata=AnnotationMetadata(tool=toolname, timestamp=timestamp),
        tokenList=TokenList(
            tokenList=[]),
        uuid=aug.next()
    )
    token_string_list = sentence_string.split()
    for i, token_string in enumerate(token_string_list):
        tokenization.tokenList.tokenList.append(Token(text=token_string,
                                                      tokenIndex=i))

    sentence = Sentence(
        textSpan=TextSpan(0, len(sentence_string)),
        tokenization=tokenization,
        uuid=aug.next()
    )

    section = Section(
        kind="SectionKind",
        sentenceList=[sentence],
        textSpan=TextSpan(0, len(sentence_string)),
        uuid=aug.next()
    )

    comm.sectionList = [section]

    return comm


class SimpleCommTempFile(object):
    '''
    Class representing a temporary file of sample concrete objects.
    Designed to facilitate testing.  Class members:

        path:           path to file
        communications: list of communications that were written to file

    Usage demo:

    >>> from concrete.util import CommunicationReader
    >>> with SimpleCommTempFile(n=3, id_fmt='temp-%d') as f:
    ...     reader = CommunicationReader(f.path)
    ...     for (orig_comm, comm_path_pair) in zip(f.communications, reader):
    ...         print orig_comm.id
    ...         print orig_comm.id == comm_path_pair[0].id
    ...         print f.path == comm_path_pair[1]
    temp-0
    True
    True
    temp-1
    True
    True
    temp-2
    True
    True
    '''

    def __init__(self, n=10, id_fmt='temp-%d',
                 sentence_fmt='Super simple sentence %d .',
                 writer_class=CommunicationWriter, suffix='.concrete'):
        '''
        Create temp file and write communications.

            n:i     number of communications to write
            id_fmt: format string used to generate communication IDs;
                    should contain one instance of %d, which will be
                    replaced by the number of the communication
            sentence_fmt: format string used to generate communication
                    IDs; should contain one instance of %d, which will
                    be replaced by the number of the communication
            writer_class: CommunicationWriter or CommunicationTGZWriter
            suffix: file path suffix (you probably want to choose this
                    to match writer_class)
        '''
        (fd, path) = tempfile.mkstemp(suffix=suffix)
        os.close(fd)
        self.path = path
        self.communications = []
        w = writer_class()
        w.open(path)
        for i in xrange(n):
            comm = create_simple_comm(id_fmt % i, sentence_fmt % i)
            self.communications.append(comm)
            w.write(comm)
        w.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if os.path.exists(self.path):
            os.remove(self.path)



#!/usr/bin/env python

# This script converts between Thrift protocol embeddings.
# Usage is:
# python convert_protocols.py \
#   --input-file /path/to/input/file
#   --output-file /path/to/output/file
#   --direction <known conversion>
#
# <known conversion> is a categorical directive, in the form
#   <input-encoding>-to-<output-encoding>
# The --input-file must be in the input format; the output file will
# be in the <output-encoding>
#
# <known conversion> must currently be either
#   * binary-to-compact: Convert the TBinaryProtocol encoded input file to a TCompactProtocol encoded output file
#   * compact-to-binary: Convert the TCompactProtocol encoded input file to a TBinaryProtocol encoded output file
#
# The known conversions are stored in a dictionary KNOWN_CONVERSIONS:
#   * keys are the conversion name, e.g., 'binary-to-compact'
#   * values are pairs of callable factory methods (callable-input-factory, callable-output-factory), e.g.,
#        (TBinaryProtocol.TBinaryProtocolFactory, TCompactProtocol.TCompactProtocolFactory)
#
# Therefore, adding to the conversions is simple: just add to the KNOWN_CONVERSIONS mapping.

import argparse
import gzip
from thrift import TSerialization
from thrift.protocol import TCompactProtocol, TBinaryProtocol, TJSONProtocol
from thrift.transport import TTransport
from concrete import Communication
import mimetypes

PROTOCOLS = {
    "binary"  : TBinaryProtocol.TBinaryProtocolFactory,
    "compact" : TCompactProtocol.TCompactProtocolFactory,
    "json"    : TJSONProtocol.TJSONProtocolFactory
    }

KNOWN_CONVERSIONS = {
    'binary-to-compact':( PROTOCOLS["binary"], PROTOCOLS["compact"]),
    'compact-to-binary':( PROTOCOLS["compact"], PROTOCOLS["binary"]),
    'compact-to-json':( PROTOCOLS["compact"], PROTOCOLS["json"]),
    'json-to-compact':( PROTOCOLS["json"], PROTOCOLS["compact"])
}

def make_parser():
    parser = argparse.ArgumentParser(description='Convert communications between protocols')
    parser.add_argument('--input-file', type=str, required=True,
                        help='input file path')
    parser.add_argument('--output-file', type=str, required=True,
                        help='output file path')
    parser.add_argument('--direction', choices = KNOWN_CONVERSIONS.keys(), required=False)
    parser.add_argument('--iprotocol', choices = sorted(PROTOCOLS.keys()), required=False)
    parser.add_argument('--oprotocol', choices = sorted(PROTOCOLS.keys()), required=False)
    return parser

def convert(input_file_path, output_file_path, input_protocol_factory, output_protocol_factory):
    """
    Convert an input file (to be read in as an input_protocol_factory type)
    to an output file (with encoding output_protocol_factory type).
    
    * input_file_path: Path to the input file (on disk)
    * output_file_path: Path the output file (on disk)
    * input_protocol_factory: Callable factory function for input encoding, e.g., TBinaryProtocol.TBinaryProtocolFactory.
    * output_protocol_factory: Callable factory function for output encoding, e.g., TCompactProtocol.TCompactProtocolFactory.
    """
    input_file = open(input_file_path,'r')
    input_bytes = input_file.read()
    output_bytes = convert_communication(input_bytes, input_protocol_factory, output_protocol_factory)
    input_file.close()
    output_file = open(output_file_path,'w')
    output_file.write(output_bytes)
    output_file.close()

def convert_communication(input_bytes, input_protocol_factory, output_protocol_factory):
    """
    Convert an input byte stream (to be read in as an input_protocol_factory type)
    to an output byte stream (with encoding output_protocol_factory type).
    
    * input_bytes: Input file byte stream
    * input_protocol_factory: Callable factory function for input encoding, e.g., TBinaryProtocol.TBinaryProtocolFactory.
    * output_protocol_factory: Callable factory function for output encoding, e.g., TCompactProtocol.TCompactProtocolFactory.
    """
    comm = Communication()
    TSerialization.deserialize(comm,
                               input_bytes,
                               protocol_factory = input_protocol_factory())
    output_bytes = TSerialization.serialize(comm, protocol_factory = output_protocol_factory())
    return output_bytes
    
if __name__ == '__main__':
    parser = make_parser()
    args = parser.parse_args()
    mimetypes.init()
    (ifile_type, ifile_encoding) = mimetypes.guess_type(args.input_file)
    (ofile_type, ofile_encoding) = mimetypes.guess_type(args.output_file)
    out_writer = None
    if args.direction is None:
        if args.iprotocol is None or args.oprotocol is None:
            print "Either --direction, or both --iprotocol and --oprotocol, must be provided"
            exit(1)
    else:
        if (not args.iprotocol is None) or (not args.oprotocol is None):
            print "Not both --direction, and either --iprotocol or --oprotocol, can be provided"
            exit(1)
    encoding_input = KNOWN_CONVERSIONS[args.direction][0] if args.iprotocol is None else PROTOCOLS[args.iprotocol]
    encoding_output = KNOWN_CONVERSIONS[args.direction][1] if args.oprotocol is None else PROTOCOLS[args.oprotocol]
    if ofile_encoding == "gzip":
        out_writer = gzip.GzipFile(args.output_file, 'wb')
    else:
        out_writer = open(args.output_file, 'w')
    if ifile_encoding == 'gzip':
        f = gzip.GzipFile(args.input_file)
        transportIn = TTransport.TFileObjectTransport(f)
        protocolIn = encoding_input().getProtocol(transportIn)
        while True:
            try:
                comm = Communication()
                comm.read(protocolIn)
                output_bytes = TSerialization.serialize(comm, protocol_factory = encoding_output())
                out_writer.write(output_bytes)
            except EOFError:
                break
        f.close()
    else:
        convert(input_file_path  = args.input_file,
                output_file_path = args.output_file,
                input_protocol_factory  = encoding_input,
                output_protocol_factory = encoding_output)
    out_writer.close()

